# Copyright 2019-2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from django.conf import settings
from django.core.management.base import BaseCommand

from django_obm import models


class Command(BaseCommand):
    help = "Creates nodes and currencies objects from initial config."

    def get_or_create_with_stdout(self, model, defaults=None, **kwargs):
        obj, created = model.objects.get_or_create(
            **kwargs, defaults=defaults or {},
        )
        if created:
            self.stdout.write(
                self.style.SUCCESS(f"{repr(obj)} created successfully.")
            )
        return obj

    def handle(self, *args, **options):
        nodes_config = getattr(settings, "OBM_NODES_INITIAL_CONFIG", [])
        for node_config in nodes_config:
            currency_config = node_config.pop("currency")
            node_config["currency"] = self.get_or_create_with_stdout(
                model=models.Currency,
                name=currency_config.pop("name"),
                defaults=currency_config,
            )
            self.get_or_create_with_stdout(model=models.Node, **node_config)



# Copyright 2019-2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TypeVar

from django.db import models
from obm import connectors, validators
from obm.sync import mixins

from django_obm import exceptions, managers

TTransaction = TypeVar("TTransaction", bound="Transaction")
TNode = TypeVar("TNode", bound="Node")


class Currency(models.Model):
    name = models.CharField(max_length=100, unique=True,)

    class Meta:
        verbose_name_plural = "currencies"

    def __str__(self):
        return self.name

    @property
    def default_node(self) -> TNode:
        default_nodes = self.nodes.filter(is_default=True)
        if len(default_nodes) == 0:
            raise exceptions.DefaultNodeDoesNotExistError(
                f"Missing default node for {self.name}"
            )
        if len(default_nodes) > 1:
            raise exceptions.TooManyDefaultNodes(
                f"Too many default nodes for {self.name}. "
                f"You can create only 1 default node."
            )
        return default_nodes.first()


class Address(models.Model):
    value = models.CharField(max_length=500,)
    currency = models.ForeignKey(
        to=Currency,
        on_delete=models.CASCADE,
        related_name="addresses",
        related_query_name="address",
    )
    password = models.CharField(max_length=500, default="")

    class Meta:
        unique_together = (("value", "currency"),)

    def __str__(self):
        return self.value


class Transaction(models.Model, mixins.TransactionMixin):
    node = models.ForeignKey(
        to="Node",
        on_delete=models.CASCADE,
        related_name="transactions",
        related_query_name="transaction",
    )
    from_address = models.ForeignKey(
        to=Address, on_delete=models.CASCADE, null=True,
    )
    to_address = models.ForeignKey(
        to=Address,
        on_delete=models.CASCADE,
        related_name="transactions",
        related_query_name="transaction",
    )
    txid = models.CharField(
        null=True, verbose_name="transaction id", max_length=500,
    )
    category = models.CharField(max_length=30)
    amount = models.DecimalField(
        max_digits=19,
        decimal_places=10,
        help_text="The transaction amount in currency.",
    )
    block_number = models.PositiveIntegerField(
        null=True, help_text="Number of block that contain the transaction.",
    )
    fee = models.DecimalField(
        null=True,
        max_digits=19,
        decimal_places=10,
        help_text=(
            "The amount of the fee in currency. This is only available for "
            "the 'send' category of transactions."
        ),
    )
    timestamp = models.PositiveIntegerField(
        null=True, help_text="Transaction creation or detection timestamp.",
    )

    objects = managers.TransactionManager()

    class Meta:
        unique_together = (("node", "txid"),)

    def __str__(self):
        return f"{self.currency}:{self.txid}"

    @property
    def confirmations_number(self) -> int:
        # TODO: To obm
        if self.block_number:
            latest_block_number = self.node.get_latest_block_number()
            self.node.close()
            return latest_block_number - self.block_number
        return 0

    @property
    def currency(self) -> Currency:
        return self.node.currency

    def send(self, subtract_fee_from_amount: bool = False):
        tx = {
            "amount": self.amount,
            "to_address": self.to_address.value,
        }
        if self.currency.name == "ethereum":
            if self.from_address:
                tx["from_address"] = self.from_address.value
                tx["password"] = self.from_address.password
        elif self.currency.name == "bitcoin" and self.fee:
            tx["fee"] = self.fee

        sent_tx = self.node.send_transaction(
            **tx, subtract_fee_from_amount=subtract_fee_from_amount,
        )
        self.node.close()
        self.txid = sent_tx["txid"]
        self.fee = sent_tx["fee"]
        self.amount = sent_tx["amount"]
        self.timestamp = sent_tx["timestamp"]
        self.save()
        return self


class Node(models.Model, mixins.ConnectorMixin):
    # TODO: Add validators
    name = models.CharField(max_length=200, unique=True,)
    currency = models.ForeignKey(
        to=Currency,
        on_delete=models.CASCADE,
        related_name="nodes",
        related_query_name="node",
    )
    is_default = models.BooleanField(
        default=True,
        help_text=(
            "If True the node will be used as default "
            "for transaction sending."
        ),
    )
    rpc_username = models.CharField(
        verbose_name="RPC username",
        max_length=200,
        help_text="Username for JSON-RPC connections.",
    )
    rpc_password = models.CharField(
        verbose_name="RPC password",
        max_length=200,
        help_text="Password for JSON-RPC connections.",
    )
    rpc_host = models.URLField(
        verbose_name="RPC host",
        default="127.0.0.1",
        help_text="Listen for JSON-RPC connections on this IP address.",
    )
    rpc_port = models.PositiveIntegerField(
        verbose_name="RPC port",
        help_text="Listen for JSON-RPC connections on this port.",
    )
    timeout = models.FloatField(
        default=connectors.DEFAULT_TIMEOUT,
        help_text="Timeout for call of node JSON RPC.",
    )

    objects = managers.NodeManager(Transaction, Address)

    class Meta:
        unique_together = (("rpc_host", "rpc_port"),)

    def __init__(self, *args, **kwargs):
        self.loop = None
        self.session = None
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):  # pylint: disable=arguments-differ
        validators.validate_node_is_supported(self.name)
        default_nodes = self.currency.nodes.filter(is_default=True).exclude(
            id=self.id
        )
        if default_nodes.count() > 0:
            raise exceptions.DefaultNodeAlreadyExists(
                f"Default node for {self.name} already exist"
            )
        super().save(*args, **kwargs)

    def delete(self, using=None, keep_parents=False):
        self.close()
        super().delete(using, keep_parents)



# Copyright 2019-2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import traceback
from datetime import datetime

from django.conf import settings
from django.core.management.base import BaseCommand

from django_obm import models


class Command(BaseCommand):
    DEFAULT_FREQUENCY = 60
    help = "Processes receipts with specified frequency."

    def add_arguments(self, parser):
        parser.add_argument(
            "--once",
            action="store_true",
            default=False,
            help="Run receipts processing only once.",
        )
        parser.add_argument(
            "--raise-errors",
            action="store_true",
            default=False,
            help="Raise any error that occur when method.",
        )
        parser.add_argument(
            "--frequency",
            action="store",
            default=getattr(
                settings,
                "OBM_COLLECT_TRANSACTION_FREQUENCY",
                self.DEFAULT_FREQUENCY,
            ),
            type=int,
            help="Raise any error that occur due execution.",
        )

    def log(self, msg, style=None):
        msg = f"{datetime.now()}:   {msg}"
        self.stdout.write(style(msg) if style else msg)

    def handle(self, *args, **options):
        frequency = options["frequency"]
        self.log(
            f"Start sync transactions proccess with "
            f"{frequency} sec. frequency"
        )

        while True:
            self.log("Run synchronization")
            try:
                result = models.Node.objects.sync_transactions()
            except Exception as exc:  # pylint: disable=broad-except
                if options["raise_errors"]:
                    raise exc
                self.log(traceback.format_exc(), style=self.style.ERROR)
            else:
                self.log("Synchronized transactions", style=self.style.SUCCESS)

            if options["once"]:
                break
            time.sleep(frequency)



import setuptools

import django_obm

EXTRAS_REQUIRE = {
    "tests": ["pytest", "pytest-django", "python-dotenv", "pytest-xdist"],
    "lint": ["pylint", "pylint-django", "mypy"],
    "docs": ["sphinx>=2.4,<3", "sphinx-rtd-theme"],
    "dev": ["tox", "rope"],
}
EXTRAS_REQUIRE["dev"] += (
    EXTRAS_REQUIRE["tests"] + EXTRAS_REQUIRE["lint"] + EXTRAS_REQUIRE["docs"]
)


def read(file_name):
    with open(file_name) as f:
        content = f.read()
    return content


setuptools.setup(
    name="django-obm",
    version=django_obm.__version__,
    packages=setuptools.find_packages(exclude=["tests*", "example*"]),
    install_requires=[
        "obm<0.1.0",
        "Django>=2.2,<4",
        "djangorestframework>=3,<4",
    ],
    extras_require=EXTRAS_REQUIRE,
    license="GNU Lesser General Public License v3 or later (LGPLv3+)",
    description=(
        "Django app that provide REST API and ORM-integrated interface for "
        "interaction with blockchains.Django app that provide REST API and "
        "ORM-integrated interface for interaction with blockchains."
    ),
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    url="https://github.com/HelloCreepy/django-obm",
    author="Alexander Polishchuk",
    author_email="apolishchuk52@gmail.com",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



# Copyright 2019-2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections

from django.db import models


class NodeManager(models.Manager):

    transaction_model = None

    def __init__(self, transaction_model=None, address_model=None):
        super().__init__()
        # Hack to avoid cyclic import
        # TODO: Point out how ti improve it
        self.transaction_model = transaction_model
        self.address_model = address_model

    def fetch_recent_transactions(self, limit: int = 50):
        """Fetches recent transactions from a blockchain."""

        def omit_info(tx):
            return {key: value for key, value in tx.items() if key != "info"}

        def to_address(value, currency):
            address, _ = self.address_model.objects.get_or_create(
                value=value, currency=currency
            )
            return address

        recent_transactions = []
        for node in self.all():
            node_recent_transactions = node.fetch_recent_transactions(limit)
            recent_transactions += [
                self.transaction_model(  # type: ignore
                    node=node,
                    from_address=to_address(
                        tx.pop("from_address"), node.currency
                    ),
                    to_address=to_address(tx.pop("to_address"), node.currency,),
                    **omit_info(tx),
                )
                for tx in node_recent_transactions
            ]
            node.close()
        return recent_transactions

    def bulk_create_recent_transactions(self, limit: int = 50) -> list:
        """Create in the database recent transactions from blockchain.

        Args:
            limit: Limit of recent transactions to create. Defaults to 50.

        Returns:
            List of created recent transactions.
        """
        recent_tx = self.fetch_recent_transactions(limit)
        return self.transaction_model.objects.bulk_create(  # type: ignore
            recent_tx, ignore_conflicts=True,
        )

    def sync_transactions(self, recent_limit: int = 50):
        def remove_duplicates(recent_txs, synchronized_txs):
            recent_txs = {tx.txid: tx for tx in recent_txs}
            synchronized_txs = {tx.txid: tx for tx in synchronized_txs}
            for txid, tx in recent_txs.items():
                if txid not in synchronized_txs:
                    synchronized_txs[txid] = tx
            return sorted(
                synchronized_txs.values(),
                key=lambda tx: tx.block_number,
                reverse=True,
            )

        recent_txs = self.bulk_create_recent_transactions(recent_limit)
        synchronized_txs = self.transaction_model.objects.sync()  # type: ignore
        return remove_duplicates(recent_txs, synchronized_txs)


class TransactionManager(models.Manager):
    def sync(self) -> list:
        """Synchronize still unconfirmed transactions with blockchain.

        Make sense to synchronize only a transaction that still have
        block_number = None. Because transaction that have already added to
        block is unchanging by blockchain design.

        Returns:
            Updated transactions list.
        """

        def update(txs, current_txs):
            txs = {tx.txid: tx for tx in txs}
            for current_tx in current_txs:
                tx = txs[current_tx["txid"]]
                tx.block_number = current_tx["block_number"]
            return list(txs.values())

        txs_by_node = collections.defaultdict(list)
        for tx in self.filter(block_number=None):
            txs_by_node[tx.node].append(tx)

        synchronized_txs = []
        for node, txs in txs_by_node.items():
            current_txs = node.fetch_in_wallet_transactions(
                [tx.txid for tx in txs]
            )
            updated_txs = update(txs, current_txs)
            synchronized_txs += updated_txs
            self.bulk_update(updated_txs, ["block_number"])
            node.close()

        return synchronized_txs
